/*
 *  Tunnel.cpp - Tunnel effect V2
 *               Written by Christian Bauer
 *				 V2 modifications by Tinic Urou
 *               Public domain
 */

#include <AppKit.h>
#include <InterfaceKit.h>
#include <KernelKit.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>


// Types
typedef unsigned char 	UBYTE;
typedef unsigned short 	UWORD;
typedef unsigned long 	ULONG;
typedef char 			BYTE;
typedef short 			WORD;
typedef long 			LONG;

// Constants

const 	ULONG 			APP_SIGNATURE = '????';
const 	LONG 			DISPLAY_X = 320;
const 	LONG 			DISPLAY_Y = 240;

// Window/Application messages

const 	LONG			MSG_REDRAW = 1;

// Global variables

		UBYTE 			DistAngBuf[DISPLAY_X * DISPLAY_Y * 2];
		UBYTE 			Texture[256 * 256];
		BYTE 			FastSin1[256];
		BYTE 			FastSin2[256];

class 					TunnelWindow;

// Application object
class Tunnel : public BApplication 
{
	public:
						Tunnel();
	virtual void 		ReadyToRun(void);
	virtual bool 		QuitRequested(void);

	private:
	
	TunnelWindow 		*the_window;
	BBitmap 			*the_bitmap;
};

struct draw_arguments
{
	UBYTE	*bits;
	UWORD	*dab;
	ULONG	line;
	ULONG	xmod;
	BYTE	angle;
};

// Window object
class TunnelWindow : public BWindow 
{
	public:
	
						TunnelWindow(int width, int height, BBitmap *bitmap);
	virtual bool 		QuitRequested(void);
	virtual void 		MessageReceived(BMessage *msg);

	private:
	
	void 				recalc(void);
	static void			draw_thread(draw_arguments *d_arg);

	BBitmap 			*the_bitmap;
	BView 				*main_view;	
	thread_id 			t1,t2;
	draw_arguments		da1,da2;
};

// Create application object and start it
int main(int argc, char **argv)
{	
	Tunnel *the_app;

	srand(system_time());

	the_app = new Tunnel();
	the_app->Run();
	delete the_app;

	return 0;
}


/*
 *  Application constructor
 */

Tunnel::Tunnel() : BApplication(APP_SIGNATURE)
{
	the_window = NULL;
	the_bitmap = NULL;
}


/*
 *  Open window and start rendering
 */

void Tunnel::ReadyToRun(void)
{
	int x, y;
	int dist;
	UBYTE *p;

	// Pre-calculate sinus to integer
	for(x=0;x<256;x++) 
	{	
		FastSin1[x]=(sin((M_PI/128)*x)+1)*127;
		FastSin2[x]=sin((M_PI/128)*x)*64;
	}
	
	// Allocate the bitmap		
	the_bitmap = new BBitmap(BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1), B_COLOR_8_BIT);

	// Pre-calculate cylindrical mapping
	p = DistAngBuf;
	for (y=0; y<DISPLAY_Y; y++)
	{
		for (x=0; x<DISPLAY_X; x++) 
		{
			dist = sqrt((x - DISPLAY_X/2) * (x - DISPLAY_X/2) + (y - DISPLAY_Y/2) * (y - DISPLAY_Y/2));
			if (dist > 16) 
			{
				*p++ = 4096 / dist;
				*p++ = atan2(x - DISPLAY_X/2, y - DISPLAY_Y/2) * 127 / M_PI;
			} 
			else 
			{
				*p++ = 0;
				*p++ = 0;
			}
		}
	}
	// Open window
	the_window = new TunnelWindow(DISPLAY_X, DISPLAY_Y, the_bitmap);

	// Start action
	the_window->PostMessage(MSG_REDRAW);
}


/*
 *  Quit requested (either by menu or by closing the window)
 */

bool Tunnel::QuitRequested(void)
{
	// Make sure that the window closes first
	if (BApplication::QuitRequested()) 
	{
		if (the_bitmap)	delete the_bitmap;
		return TRUE;
	}
	return FALSE;
}


/*
 *  Window constructor
 */

TunnelWindow::TunnelWindow(int width, int height, BBitmap *bitmap) : BWindow(BRect(0, 0, width-1, height-1), "Tunnel", B_TITLED_WINDOW, B_NOT_RESIZABLE | B_NOT_ZOOMABLE)
{
	the_bitmap = bitmap;
	
	ULONG xmod = bitmap->BytesPerRow();

	// Preset most used pointers for threadfunction to avoid globalisation
	
	da2.xmod=da1.xmod=xmod;
	da2.bits=(UBYTE *)the_bitmap->Bits();
	da1.bits=((UBYTE *)the_bitmap->Bits()+((DISPLAY_Y/2)*xmod));
	da2.dab=(UWORD *)DistAngBuf;
	da1.dab=(UWORD *)(DistAngBuf+(DISPLAY_Y*DISPLAY_X));

	// Move window to right position
	MoveTo(80, 60);

	// Create bitmap view
	Lock();
	main_view = new BView(BRect(0, 0, DISPLAY_X-1, DISPLAY_Y-1),"", B_FOLLOW_NONE,NULL);
	AddChild(main_view);
	Unlock();

	// Show the window
	Show();
}


/*
 *  Closing the window quits the program
 */

bool TunnelWindow::QuitRequested(void)
{
	be_app->PostMessage(B_QUIT_REQUESTED);
	return TRUE;
}


/*
 *  Handles redraw messages
 */

void TunnelWindow::MessageReceived(BMessage *msg)
{
	switch (msg->what) {
		case MSG_REDRAW:  // Redraw bitmap
			Lock();
			main_view->DrawBitmapAsync(the_bitmap);
			recalc();
			main_view->Sync();
			PostMessage(MSG_REDRAW);
			Unlock();
			break;
		default:
			BWindow::MessageReceived(msg);
	}
}


/*
 *  Main calculation
 */

void TunnelWindow::draw_thread(draw_arguments *d_arg)
{
	// Speeds up access!!
	UBYTE	*p		= d_arg->bits;
	UWORD	*q 		= d_arg->dab;
	UWORD	line	= d_arg->line;
	ULONG	xmod	= d_arg->xmod;
	UWORD	angle	= d_arg->angle;

	for (ULONG y=(DISPLAY_Y/2);y;y--,p+=xmod)
	{
		for (ULONG x=0; x<DISPLAY_X; x++, q++)
		{
			if (*q) 
			{
				UWORD index=*q;
				index = ((index+line)&0xff00)|((index+angle)&0x00ff);
				p[x] = Texture[index];
			} 
			else 
			{
				p[x] = 0;	
			}
		}
	}
}

void TunnelWindow::recalc(void)
{
	static UWORD line=0;
	static LONG red_phase=0, green_phase=0, blue_phase=0, returnvalue=0;
	
	// Setup arguments for threadfunctions
	
	da2.angle=da1.angle=FastSin2[line>>8];
	da2.line=da1.line=line;

	// Draw tunnel using two threads
	
	resume_thread(t1=spawn_thread((thread_entry)draw_thread,"Tunnel Draw Thread 1",B_NORMAL_PRIORITY,&da1));
	resume_thread(t2=spawn_thread((thread_entry)draw_thread,"Tunnel Draw Thread 2",B_NORMAL_PRIORITY,&da2));
	wait_for_thread(t1,&returnvalue);
	wait_for_thread(t2,&returnvalue);

	// Add new texture data

	UBYTE *p = Texture + line;
	line 	+= 0x400;

	for (ULONG y=0; y<4; y++) 
	{
		for (ULONG x=0; x<256; x++) 
		{
			*p++ = index_for_color
			(
				FastSin1[(UBYTE)(x+red_phase)  ],
				FastSin1[(UBYTE)(x+green_phase)],
				FastSin1[(UBYTE)(x+blue_phase) ]		
			);
		}
		red_phase 	+= (rand()-(RAND_MAX/2))/(RAND_MAX/24);
		green_phase += (rand()-(RAND_MAX/2))/(RAND_MAX/24);
		blue_phase 	+= (rand()-(RAND_MAX/2))/(RAND_MAX/24);
	}
}
